/****************************************************************************
  Author  :  Asger Hobolth and Ruriko Yoshida
  Code    :  The algorithm, the estimation maximization generalized neighbor-
             joining method, described in AB2005.
  Function:  Estimating a phylogenetic tree under the JC model.
             
  Date    :  Dec 9th., 2005
  Update  :  Dec 16th., 2005

  Default invocation:
  ./JC-Tree numItr <filename>
  where numItr is the number of iterations and <filename> is the input file
  firmatted in fasta format.

  Note: This will reconstruct a tree under the JC model.

  Note: The number of iteration must come to the first arguement.

******************************************************************************/

#include "TreeNode.h"
#include "Newick.h"
#include <lapack++.h>
#include <lapack.h>
#include <lafnames.h>
#include <iostream>
#include "Felsenstein.h"
#include "Mjoin.h"
#include "Seq.h"

using namespace std;

long double JC_tree(int** SeqMat, int** PhlMat, LaSymmMatDouble RateMat, 
	     LaVectorDouble Frq, LaVectorDouble& BrLen, int numSt, int numLvs, 
	     int numNds, int nIter, int SeqLen){
  long double likVal = 0.0, logL, tmp = 0.0, holding[numSt * numSt];
  LaSymmMatDouble Q(numSt, numSt);
  LaGenMatDouble tmpCndProbMat(numNds + numLvs - 1, numSt * numSt), 
    CndProbMat(numNds + numLvs - 1, numSt * numSt);
  int Obs[numLvs];

  //  logL = new long double[nIter];

  for(int itr = 0; itr < nIter; itr++){
    cout << "Iteration: " << itr + 1 << "...." << endl;
    for(int i = 0; i < (numNds + numLvs - 1); i++)
      for(int j = 0; j < (numSt * numSt); j++)
	CndProbMat(i, j) = 0.0;
	
    logL = 0.0;
    tmp = 0.0;

    // cout << SeqLen << endl;
    for(int j = 0; j < SeqLen; j++){
      tmp = 0;
      for(int i = 0;  i < numLvs; i++)
	Obs[i] = SeqMat[i][j];

      for(int i = 0; i < (numNds + numLvs - 1); i++)
	for(int j = 0; j < (numSt * numSt); j++)
	  tmpCndProbMat(i, j) = 0.0;
      
      FlsnFct(PhlMat, RateMat, Frq, BrLen, Obs, numSt, numLvs, numNds, tmpCndProbMat, tmp);
      logL += tmp;     
      // cout << tmpCndProbMat << endl;

    for(int i = 0; i < (numNds + numLvs - 1); i++)
      for(int j = 0; j < (numSt * numSt); j++)
	CndProbMat(i, j) += tmpCndProbMat(i, j);
    }

    for(int nd = 0; nd < (numNds + numLvs - 1); nd++){

      for(int i = 0; i < numSt; i++)
	for(int j = 0; j < numSt; j++) Q(i, j) = 0.0;

      for(int i = 0; i < numSt; i++)
	for(int j = 0; j < numSt; j++)
	  Q(i, j) = BrLen(nd) * RateMat(i, j);
      
      for(int k = 0; k < (numSt * numSt); k++) holding[k] = 0.0;

      for(int i = 0; i < (numSt * numSt); i++)
	for(int j = 0; j < (numSt * numSt); j++)
	  holding[i] += CndProbMat(nd, j) * (Transposition(CndStatFct(Frq, Q, numSt), numSt*numSt))(j, i);
      
      //  cout << CndProbMat<<endl<< (Transposition(CndStatFct(Frq, Q, numSt), numSt*numSt))<< endl << CndStatFct(Frq, Q, numSt) << endl; exit(0);
//       cout << nd << endl;
//       for(int k = 0; k < (numSt * numSt); k++) cout << holding[k] << " ";
//       cout << endl;
      BrLen(nd) = holding[4];
      
      for(int i = 5; i < (numSt * numSt); i++)
	BrLen(nd) += holding[i];
      BrLen(nd) /= SeqLen;
    }
    cout << "logL = " << logL << endl;
    cout << "Branch length: ";
    for(int i = 0; i < (numNds + numLvs - 1); i++)  
      cout << BrLen(i) << " ";
    cout << endl;
  }// long forloop ends for itr
  
  return logL;
}

void runJC_Tree(int argc, char **argv){
  /* Precision for printing.... */
  cout.precision(4);
  cout.setf(ios::scientific, ios::floatfield);

  /* Declearing variables.... */
  int numSt = 4, numNds = 4, numLvs = 5, ** PhylMat, **SeqMat, nIter = 0;
  int Obs[numLvs], SeqLen = 0;
  LaVectorDouble BrLenV(numLvs + numNds - 1), BrLenV2;
  double logL = 0.0;
  LaSymmMatDouble RateMat(numSt, numSt), Q(numSt, numSt);
  LaGenMatDouble CndProbMat(numNds + numLvs - 1, numSt * numSt);
  char invocation[200], filename[127], Newick[127];
  Alinements * InputData;
  LaVectorDouble FreqVec(numSt);
  
  /* Reading input files and arguements.... */
  if(argc < 3){
    cerr << "Please enter an input file name!" << endl;
    exit(0);
  }
  strcpy(filename, argv[argc-1]);
  strcpy(Newick, "no");
  
//   strcpy(invocation, "invocation: ");
//   for(int f = 0; f < argc; f++){
//     strcat(invocation, argv[f]);
//     strcat(invocation, " ");
//     if(strncmp(argv[f], "-r", 2) == 0)    strcpy(Newick, "yes");
//   }
  
  nIter = atoi(argv[1]);
  
  /* Setting the JC model..... */
  for(int i = 0; i < numSt; i++) FreqVec(i) = 0.25;
  
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      RateMat(i, j) = ((double)1)/((double)3);
      Q(i, j) = ((double)1)/((double)6);
    }
  for(int i = 0; i < numSt; i++) {RateMat(i, i) = -1.0; Q(i, i) = -0.5;}
  
  /* Initializing variables...... */  
  PhylMat = new int*[numLvs + numNds];
  for(int i = 0; i < (numLvs + numNds); i++)
    PhylMat[i] = new int[4];
  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < 4; j++) 
      PhylMat[i][j] = 0;
  
  numLvs = 0;
  
  /* Writing tree matrix from DNA sequences using MJOIN to get initial Branchlengths..... */
  PhylMat = writeTreeMatrix(filename, Newick, numLvs, BrLenV2, &InputData);
  //cout << InputData[0].DNA.length() << endl;
  numNds = numLvs - 1; 
  int dim = (2 * numLvs) - 1; 
//   cout << dim << endl;
//     for(int i = 0; i < dim; i++){
//       for(int j = 0; j < 4; j++)
//         cout << PhylMat[i][j] << " ";
//       cout << endl;
//   }
  
  /* Converting letters to numbers.... */
  SeqMat = DNAMat(InputData, numLvs, SeqLen);
  //cout.precision(16);
  //cout << fixed << BrLenV2 << endl;
  JC_tree(SeqMat, PhylMat, RateMat, FreqVec, BrLenV2, numSt, numLvs, 
	  numNds, nIter, SeqLen);
}

void Test(int argc, char **argv)
{ 
  cout.precision(4);
  cout.setf(ios::scientific, ios::floatfield);
  int numSt = 4, numNds = 4, numLvs = 5, ** PhylMat, **SeqMat;
  int Obs[numLvs], **PhlMat, SeqLen;
  LaVectorDouble BrLenV(numLvs + numNds - 1), BrLenV2;
  long double logL = 0.0;
  LaSymmMatDouble RateMat(numSt, numSt), Q(numSt, numSt);
  LaGenMatDouble CndProbMat(numNds + numLvs - 1, numSt * numSt);
  char invocation[200], filename[127], Newick[127];
  Alinements * InputData;

  LaVectorDouble FreqVec(numSt);
  
  if(argc < 2){
    cerr << "Please enter an input file name!" << endl;
    exit(0);
  }
  strcpy(filename, argv[argc-1]);
  strcpy(Newick, "no");

  strcpy(invocation, "invocation: ");
  for(int f = 0; f < argc; f++){
    strcat(invocation, argv[f]);
    strcat(invocation, " ");
    if(strncmp(argv[f], "-r", 2) == 0)    strcpy(Newick, "yes");
  }

  for(int i = 0; i < numSt; i++) FreqVec(i) = 0.25;
  
  PhlMat = new int*[numLvs + numNds];
  for(int i = 0; i < (numLvs + numNds); i++)
    PhlMat[i] = new int[4];
  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < 4; j++) 
      PhlMat[i][j] = 0;
  
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      RateMat(i, j) = ((double)1)/((double)3);
      Q(i, j) = ((double)1)/((double)6);
    }
  for(int i = 0; i < numSt; i++) {RateMat(i, i) = -1.0; Q(i, i) = -0.5;}
  
  PhlMat[0][0] = 1; PhlMat[1][0] = 1; PhlMat[2][0] = 1; PhlMat[3][0] = 1;
  PhlMat[4][0] = 1;
  PhlMat[0][3] = 6; PhlMat[1][3] = 6; PhlMat[2][3] = 7; PhlMat[3][3] = 7;
  PhlMat[4][3] = 8;
  
  PhlMat[5][1] = 1; PhlMat[5][2] = 2; PhlMat[5][3] = 9; PhlMat[6][1] = 3;
  PhlMat[6][2] = 4; PhlMat[6][3] = 8;
  PhlMat[7][1] = 7; PhlMat[7][2] = 5; PhlMat[7][3] = 9;
  PhlMat[8][1] = 6; PhlMat[8][2] = 8; PhlMat[8][3] = 0;
  
  Obs[0] = 1;  Obs[1] = 2; Obs[2] = 2; Obs[3] = 2; Obs[4] = 4;
  BrLenV(0) = 0.1;  BrLenV(1) = 0.1;  BrLenV(2) = 0.5;  BrLenV(3) = 0.5;
  BrLenV(4) = 0.1;  BrLenV(5) = 0.7;  BrLenV(6) = 0.1;  BrLenV(7) = 2.0;
  
  FlsnFct(PhlMat, RateMat, FreqVec, BrLenV, Obs, numSt, numLvs, numNds, CndProbMat, logL);
  for(int i = 0; i < 9; i++){
    for(int j = 0; j < 4; j++)
      cout << PhlMat[i][j] << " ";
    cout << endl;
  }
  cout << CndProbMat << endl;
  cout << logL << endl;
  double sum = 0.0;
  for(int i = 0; i < 16; i++) sum += CndProbMat(i, 0);
  cout << sum << endl;
  // cout << CndStatFct(FreqVec, Q, numSt) << endl;

  numLvs = 0;
  PhylMat = writeTreeMatrix(filename, Newick, numLvs, BrLenV2, &InputData);
  //cout << InputData[0].DNA.length() << endl;
  numNds = numLvs - 1; 
   int dim = (2 * numLvs) - 1; 
//   for(int i = 0; i < dim; i++){
//     for(int j = 0; j < 4; j++)
//       cout << PhylMat[i][j] << " ";
//     cout << endl;
//   }

   for(int i = 0; i < numSt; i++)
     for(int j = 0; j < numSt; j++)
       Q(i, j) =  RateMat(i, j);

   cout << CndStatFct(FreqVec, Q, numSt) << endl;
   sum = 0.0;
   for(int i = 0; i < 16; i++) sum += (CndStatFct(FreqVec, Q, numSt))(i, 0);
   cout << sum << endl;
   SeqMat = DNAMat(InputData, numLvs, SeqLen);
  // cout << endl << BrLenV2 << endl;
}

// int main(int argc, char **argv)
// {
//   runJC_Tree(argc, argv);
//   //Test(argc, argv);
//   return 0;
// }

void runJC_Tree(int nIter, int numLvs, int numSt, int** PhylMat,
		LaVectorDouble BrLenV2, int** SeqMat, int SeqLen){
  /* Precision for printing.... */
  cout.precision(4);
  cout.setf(ios::scientific, ios::floatfield);

  /* Declearing variables.... */
  int numNds = numLvs - 1;
  int Obs[numLvs];
  //LaVectorDouble BrLenV(numLvs + numNds - 1);
  double logL = 0.0;
  LaSymmMatDouble RateMat(numSt, numSt), Q(numSt, numSt);
  LaGenMatDouble CndProbMat(numNds + numLvs - 1, numSt * numSt);
  char invocation[200], filename[127], Newick[127];
  Alinements * InputData;
  LaVectorDouble FreqVec(numSt);
  
  /* Setting the JC model..... */
  for(int i = 0; i < numSt; i++) FreqVec(i) = 0.25;
  
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      RateMat(i, j) = ((double)1)/((double)3);
      Q(i, j) = ((double)1)/((double)6);
    }
  for(int i = 0; i < numSt; i++) {RateMat(i, i) = -1.0; Q(i, i) = -0.5;}
  
  /* Writing tree matrix from DNA sequences using MJOIN to get initial Branchlengths..... */
   //cout << InputData[0].DNA.length() << endl;
  numNds = numLvs - 1; 
  int dim = (2 * numLvs) - 1; 
//   cout << dim << endl;
//     for(int i = 0; i < dim; i++){
//       for(int j = 0; j < 4; j++)
//         cout << PhylMat[i][j] << " ";
//       cout << endl;
//   }
  
  /* Converting letters to numbers.... */
  //  SeqMat = DNAMat(InputData, numLvs, SeqLen);
  //cout.precision(16);
  //cout << fixed << BrLenV2 << endl;
  JC_tree(SeqMat, PhylMat, RateMat, FreqVec, BrLenV2, numSt, numLvs, 
	  numNds, nIter, SeqLen);
}
